
import inox._
import inox.trees.{forall => _, _}
import inox.trees.dsl._

import welder._

object FoldContent {

  def main2(args: Array[String]): Unit = {

    val foldID = FreshIdentifier("fold")
    val contentID = FreshIdentifier("content")
    val removeID = FreshIdentifier("remove")

    val list = FreshIdentifier("List")
    val cons = FreshIdentifier("Cons")
    val nil = FreshIdentifier("Nil")

    val head = FreshIdentifier("head")
    val tail = FreshIdentifier("tail")

    val listADT = mkSort(list)("A")(Seq(cons, nil))
    val consADT = mkConstructor(cons)("A")(Some(list)) { case Seq(aT) =>
      Seq(ValDef(head, aT), ValDef(tail, T(list)(aT)))
    }
    val nilADT = mkConstructor(nil)("A")(Some(list))(_ => Nil)

    val foldFunction = mkFunDef(foldID)("A") { case Seq(aT) => (
      Seq("f" :: ((aT, aT) =>: aT), "z" :: aT, "l" :: T(list)(aT)), aT, { case Seq(f, z, l) =>
        if_ (l.isInstOf(T(cons)(aT))) {
          E(foldID)(aT)(f, f(z, l.asInstOf(T(cons)(aT)).getField(head)), l.asInstOf(T(cons)(aT)).getField(tail))
        } else_ {
          z
        }
      })
    }

    val contentFunction = mkFunDef(contentID)("A") { case Seq(aT) => (
      Seq("l" :: T(list)(aT)), SetType(aT), { case Seq(l) =>
        if_ (l.isInstOf(T(cons)(aT))) {
          SetAdd(E(contentID)(aT)(l.asInstOf(T(cons)(aT)).getField(tail)), l.asInstOf(T(cons)(aT)).getField(head))
        } else_ {
          FiniteSet(Seq.empty, aT)
        }
      })
    }

    val removeFunction = mkFunDef(removeID)("A") { case Seq(aT) => (
      Seq("l" :: T(list)(aT), "e" :: aT), T(list)(aT), { case Seq(l, e) =>
        if_ (l.isInstOf(T(cons)(aT))) {
          if_ (l.asInstOf(T(cons)(aT)).getField(head) === e) {
            E(removeID)(aT)(l.asInstOf(T(cons)(aT)).getField(tail), e)
          } else_ {
            T(cons)(aT)(l.asInstOf(T(cons)(aT)).getField(head), E(removeID)(aT)(l.asInstOf(T(cons)(aT)).getField(tail), e))
          }
        } else_ {
          T(nil)(aT)()
        }
      })
    }

    val symbols = NoSymbols
      .withADTs(Seq(listADT, consADT, nilADT))
      .withFunctions(Seq(foldFunction, contentFunction, removeFunction))

    val program = InoxProgram(Context.empty, symbols)

    val theory = theoryOf(program)
    import theory._

    val A = TypeParameter.fresh("A")

    def fold(f: Expr, z: Expr, l: Expr) = E(foldID)(A)(f, z, l)
    def content(e: Expr) = E(contentID)(A)(e)
    def remove(l: Expr, e: Expr) = E(removeID)(A)(l, e)

    val ListA = T(list)(A)
    val ConsA = T(cons)(A)
    val NilA = T(nil)(A)

    def isAssociative(f: Expr): Expr = forall("x" :: A, "y" :: A, "z" :: A)((x, y, z) => f(x, f(y, z)) === f(f(x, y), z))
    def isCommutative(f: Expr): Expr = forall("x" :: A, "y" :: A)((x, y) => f(x, y) === f(y, x))
    def isIdempotent(f: Expr): Expr = forall("x" :: A, "y" :: A)((x, y) => f(x, y) === f(f(x, y), y))

    // forall((l: List[A], e: A) => content(remove(l, e)) == content(l) -- Set(e))
    val removeOnContent = structuralInduction((l: Expr) => {
      forall("e" :: A)(e => content(remove(l, e)) === content(l) -- FiniteSet(Seq(e), A))
    }, "l" :: ListA) { case (ihs, goal) =>
      ihs.expression match {
        case C(`cons`, h, t) => ihs.hypothesis(t)
        case C(`nil`) => trivial
      }
    }

    // forall((l: List[A], e: A) => !(content(l) contains e) ==> (remove(l, e) == l))
    val notContainsRemoveNoop = structuralInduction((l: Expr) => {
      forall("e" :: A)(e => !(content(l) contains e) ==> (remove(l, e) === l))
    }, "l" :: ListA) { case (ihs, goal) =>
      ihs.expression match {
        case C(`cons`, h, t) => ihs.hypothesis(t)
        case C(`nil`) => trivial
      }
    }

    // forall { (f: (A,A) => A, l: List[A], z: A, e: A) =>
    //   isIdempotent(f) ==> (fold(f, f(z, e), l) == fold(f, f(z, e), Cons(e, l)))
    // }
    val foldIdempotentHeadNoop = forallI("f" :: ((A, A) =>: A)) { f =>
      implI(isIdempotent(f)) { isIdem =>
        structuralInduction((l: Expr) => {
          forall("z" :: A, "e" :: A) { (z, e) =>
            fold(f, f(z, e), l) === fold(f, f(z, e), ConsA(e, l))
          }
        }, "l" :: ListA) { case (ihs, goal) =>
          andI(isIdem, ihs.expression match {
            case C(`cons`, h, t) => ihs.hypothesis(t)
            case C(`nil`) => truth
          })
        }
      }
    }

    // forall { (f: (A,A) => A, l: List[A], z: A, e: A) =>
    //   (
    //     isAssociative(f)        &&
    //     isCommutative(f)        &&
    //     isIdempotent(f)         &&
    //     (content(l) contains e)
    //   ) ==> (
    //     fold(f, z, l) == fold(f, z, Cons(e, remove(l, e)))
    //   )
    // }
    val lemmaFoldReorder = forallI("f" :: ((A, A) =>: A)) { f =>

      // forall((l: List[A], z: A, e: A) => !(content(l) contains e) ==> (fold(f, z, l) == fold(f, z, remove(l, e))))
      val notContainsEqFold = forallI("l" :: ListA) { l =>
        forallI("z" :: A, "e" :: A) { (z, e) =>
          prove(!(content(l) contains e) ==> (
            fold(f, z, l) === fold(f, z, remove(l, e))
          ), notContainsRemoveNoop)
        }
      }
 
      implI(and(isAssociative(f), isCommutative(f), isIdempotent(f))) { fconds =>
        val Seq(isAssoc, isComm, isIdem) = andE(fconds): Seq[Theorem]

        def lhs(l: Expr, z: Expr, e: Expr) = fold(f, z, l)
        def rhs(l: Expr, z: Expr, e: Expr) = fold(f, z, ConsA(e, remove(l, e)))
        def property(l: Expr, z: Expr, e: Expr) = lhs(l, z, e) === rhs(l, z, e)

        structuralInduction(
          (l: Expr) => forall("z" :: A, "e" :: A)((z, e) => (content(l) contains e) ==> property(l, z, e)),
          "l" :: ListA
        ) { case (ihs, goal) =>
          ihs.expression match {
            case C(`cons`, h, t) =>
              forallI("z" :: A, "e" :: A) { (z, e) =>
                implI(content(ihs.expression) contains e) { inExpr =>

                  // !(content(t) contains e) ==> (fold(f, z, l) == fold(f, z, remove(l, e)))
                  val lemNotContains = implI(!(content(t) contains e)) { notInTail =>
                    val hEqE = prove(h === e, inExpr, notInTail)

                    lhs(ihs.expression, z, e)                                      ==|
                                                                                hEqE |
                    fold(f, z, ConsA(e, t))                                        ==|
                      andI(notInTail, forallE(forallE(notContainsRemoveNoop)(t))(e)) |
                    fold(f, z, ConsA(e, remove(t, e)))                             ==|
                                                                                hEqE |
                    rhs(ihs.expression, z, e)
                  }

                  // (content(t) contains e) ==> (fold(f, z, l) == fold(f, z, remove(l, e)))
                  val lemContains = implI(content(t) contains e) { inTail =>

                    val lemEq = implI(e === h) { eEqH =>
                      val removeEqTail = prove(remove(ihs.expression, e) === remove(t, e), eEqH)

                      val foldIdem = implE(forallE(foldIdempotentHeadNoop)(f))(_.by(isIdem))

                      lhs(ihs.expression, z, e)                                   ==|
                                                                            trivial |
                      fold(f, f(z, h), t)                                         ==|
                                                                               eEqH |
                      fold(f, f(z, e), t)                                         ==|
                        implE(forallE(ihs.hypothesis(t))(f(z, e), e))(_.by(inTail)) |
                      fold(f, f(z, e), ConsA(e, remove(t, e)))                    ==|
                         andI(eEqH, forallE(forallE(foldIdem)(remove(t, e)))(z, e)) |
                      fold(f, f(z, e), remove(t, e))                              ==|
                                                                       removeEqTail |
                      fold(f, f(z, e), remove(ihs.expression, e))                 ==|
                                                                               eEqH |
                      rhs(ihs.expression, z, e)
                    }

                    val lemNeq = implI(!(e === h)) { eNeqH =>
                      val removeCommutesTail = (
                        remove(ihs.expression, e) ==| eNeqH |
                        ConsA(h, remove(t, e))
                      )

                      val fProp = (
                        f(f(z, h), e)             ==|
                          forallE(isAssoc)(z, h, e) |
                        f(z, f(h, e))             ==|
                              forallE(isComm)(e, h) |
                        f(z, f(e, h))             ==|
                          forallE(isAssoc)(z, e, h) |
                        f(f(z, e), h)
                      )

                      lhs(ihs.expression, z, e)                                   ==|
                                                                            trivial |
                      fold(f, f(z, h), t)                                         ==|
                        implE(forallE(ihs.hypothesis(t))(f(z, h), e))(_.by(inTail)) |
                      fold(f, f(z, h), ConsA(e, remove(t, e)))                    ==|
                                                                            trivial |
                      fold(f, z, ConsA(h, ConsA(e, remove(t, e))))                ==|
                                                                              fProp |
                      fold(f, z, ConsA(e, ConsA(h, remove(t, e))))                ==|
                                                                 removeCommutesTail |
                      rhs(ihs.expression, z, e)
                    }

                    andI(lemEq, lemNeq)
                  }

                  andI(lemNotContains, lemContains)
                }
              }
            case C(`nil`) => trivial
          }
        }
      }
    }

    val theorem = forallI("f" :: ((A, A) =>: A)) { f =>
      implI(and(isAssociative(f), isCommutative(f), isIdempotent(f))) { fconds =>
        val Seq(isAssoc, isComm, isIdem) = andE(fconds): Seq[Theorem]

        def property(l: Expr) = forall("z" :: A, "l2" :: ListA) {
          (z, l2) => (content(l) === content(l2)) ==> (fold(f, z, l) === fold(f, z, l2))
        }

        structuralInduction(property(_), "l" :: ListA) { case (ihs, goal) =>

          ihs.expression match {
            case C(`cons`, h, t) => forallI("z" :: A, "l2" :: ListA) { (z, l2) =>
              implI(content(ihs.expression) === content(l2)) { contentEq =>
                val l2NonEmpty = prove(!(l2 === NilA()), contentEq)

                val idempotentOnContent = prove(
                  (content(t) contains h) ==> (fold(f, z, t) === fold(f, z, ConsA(h, t))),
                  forallE(ihs.hypothesis(t))(z, ConsA(h, t).asInstOf(ListA))
                )

                val simpleCase = prove((
                  l2.asInstOf(ConsA).getField(head) === h &&
                  content(t) === content(l2.asInstOf(ConsA).getField(tail))
                ) ==> (
                  fold(f, z, ihs.expression) === fold(f, z, l2)
                ), forallE(ihs.hypothesis(t))(f(z, h), l2.asInstOf(ConsA).getField(tail)), l2NonEmpty)

                val tailCase: Theorem = {
                  val lem = prove(
                    (content(t) === content(l2)) ==> (content(t) contains h) &&
                    (content(t) === content(l2)) ==> (content(t) === content(ConsA(h, t))),
                    contentEq
                  )

                  prove((
                    content(t) === content(l2)
                  ) ==> (
                    fold(f, z, ihs.expression) === fold(f, z, l2)
                  ), idempotentOnContent, lem, forallE(ihs.hypothesis(t))(z, l2), forallE(ihs.hypothesis(t))(z, ConsA(h, t).asInstOf(ListA)), l2NonEmpty)
                }

                val lastCase = implI(!(content(t) === content(l2))) { contentTailNotEq =>

                  val contentEqRemove = prove(
                    content(t) === content(remove(l2, h)),
                    contentEq,
                    contentTailNotEq,
                    forallE(forallE(removeOnContent)(l2))(h)
                  )

                  val foldEqRemove = implE(forallE(ihs.hypothesis(t))(f(z, h), remove(l2, h)))(_.by(contentEqRemove))

                  val lemmaReorder = implE(forallE(forallE(implE(forallE(lemmaFoldReorder)(f))(_.by(fconds)))(l2))(z, h))(_.by(contentEq))

                  val lem = (
                    fold(f, z, ihs.expression) ==|
                    trivial |
                    fold(f, f(z, h), t) ==|
                    foldEqRemove |
                    fold(f, f(z, h), remove(l2, h)) ==|
                    trivial |
                    fold(f, z, ConsA(h, remove(l2, h))) ==|
                    lemmaReorder |
                    fold(f, z, l2)
                  )

                  lem
                }

                andI(simpleCase, tailCase, lastCase)
              }
            }

            case C(`nil`) => trivial
          }
        }
      }
    }

    println(theorem)
  }
}
